@@ -2,7 +2,6 @@
 || This file is part of Pike. For copyright information see COPYRIGHT.
 || Pike is distributed under GPL, LGPL and MPL. See the file COPYING
 || for more information.
-|| $Id: 61330fec59c0c2b74fec8e5645f7251afce12930 $
 */
 
 #include "global.h"
@@ -36,19 +35,13 @@
 #include "pike_rusage.h"
 #include "module_support.h"
 #include "opcodes.h"
-
-#ifdef AUTO_BIGNUM
 #include "bignum.h"
-#endif
 
 #include "pike_embed.h"
 
 #ifdef LIBPIKE
 #if defined(HAVE_DLOPEN) && defined(HAVE_DLFCN_H)
 #include <dlfcn.h>
-#elif !defined(USE_DLL) && defined(USE_MY_WIN32_DLOPEN)
-#include "pike_dlfcn.h"
-#else
 #undef LIBPIKE
 #endif
 #endif
@@ -98,16 +91,21 @@
 #define MAXPATHLEN 32768
 #endif
 
-static char master_location[MAXPATHLEN * 2] = MASTER_COOKIE;
+static const char _master_location[MAXPATHLEN+CONSTANT_STRLEN(MASTER_COOKIE)] = MASTER_COOKIE;
+static const char *master_file_location = _master_location + CONSTANT_STRLEN(MASTER_COOKIE);
 
 static void set_master(const char *file)
 {
-  if (strlen(file) >= MAXPATHLEN*2 - CONSTANT_STRLEN(MASTER_COOKIE)) {
-    fprintf(stderr, "Too long path to master: \"%s\" (limit:%"PRINTPTRDIFFT"d)\n",
-	    file, MAXPATHLEN*2 - CONSTANT_STRLEN(MASTER_COOKIE));
-    exit(1);
-  }
-  strcpy(master_location + CONSTANT_STRLEN(MASTER_COOKIE), file);
+  if( master_file_location != _master_location+CONSTANT_STRLEN(MASTER_COOKIE))
+    free((void*)master_file_location);
+#ifdef DEBUG_MALLOC
+#undef strdup /* We can't use dmalloc strdup before pike_memory is initialized. */
+#endif /* DEBUG_MALLOC */
+#if defined(__NT__) && !defined(strdup)
+  /* strdup() is broken on WIN32, but _strdup() isn't... */
+#define strdup _strdup
+#endif /* __NT__ */
+  master_file_location = strdup( file );
 }
 
 #ifdef __NT__
@@ -143,13 +141,10 @@ static void get_master_key(HKEY cat)
 
 static void set_default_master(const char *bin_name)
 {
-  char *mp = master_location + CONSTANT_STRLEN (MASTER_COOKIE);
-
-#ifdef HAVE_GETENV
-  if(!*mp && getenv("PIKE_MASTER")) {
+#define mp master_file_location
+  if(getenv("PIKE_MASTER")) {
     set_master(getenv("PIKE_MASTER"));
   }
-#endif
 
 #ifdef __NT__
   if(!*mp) get_master_key(HKEY_CURRENT_USER);
@@ -158,11 +153,13 @@ static void set_default_master(const char *bin_name)
 
   if(!*mp && strncmp(DEFAULT_MASTER, "NONE/", 5))
   {
-    SNPRINTF (mp, sizeof (master_location) - CONSTANT_STRLEN (MASTER_COOKIE),
+    char tmp[CONSTANT_STRLEN( DEFAULT_MASTER ) + 10 + 10 + 10];
+    SNPRINTF (tmp, sizeof(tmp)-1,
 	      DEFAULT_MASTER,
 	      PIKE_MAJOR_VERSION,
 	      PIKE_MINOR_VERSION,
 	      PIKE_BUILD_VERSION);
+    set_master( tmp );
   }
 
 #ifdef __NT__
@@ -176,27 +173,28 @@ static void set_default_master(const char *bin_name)
       char *p = strrchr (exepath, '\\');
       if (p) *p = 0;
       SNPRINTF (tmp, sizeof (tmp), "%s/%s", exepath, mp);
-      strncpy (mp, tmp,
-	       sizeof (master_location) - CONSTANT_STRLEN (MASTER_COOKIE));
+      set_master( tmp );
     }
   }
 #else
   if (!*mp) {
     /* Attempt to find a master via the path to the binary. */
     /* Note: We assume that MAXPATHLEN is > 18 characters. */
-    if (strlen(bin_name) < (2*MAXPATHLEN -
-			    CONSTANT_STRLEN(MASTER_COOKIE "master.pike"))) {
+    if (strlen(bin_name) + CONSTANT_STRLEN("master.pike") < MAXPATHLEN) {
+      char tmp[MAXPATHLEN];
       char *p;
-      strcpy(mp, bin_name);
-      p = strrchr(mp, '/');
-      if (!p) p = mp;
+      strcpy(tmp, bin_name);
+      p = strrchr(tmp, '/');
+      if (!p) p = tmp;
       else p++;
       strcpy(p, "master.pike");
+      set_master( tmp );
     }
   }
 #endif
 
   TRACE((stderr, "Default master at \"%s\"...\n", mp));
+#undef mp
 }
 
 #ifdef LIBPIKE
@@ -277,8 +275,7 @@ static void find_lib_dir(int argc, char **argv)
   {
     char *p;
     char *dir;
-    memcpy(libpike_file, master_location + CONSTANT_STRLEN(MASTER_COOKIE),
-	   sizeof(master_location) - CONSTANT_STRLEN(MASTER_COOKIE));
+    strcpy(libpike_file, master_file_location);
     for (p = dir = libpike_file; *p; p++) {
       if ((*p == '/')
 #ifdef __NT__
@@ -304,9 +301,7 @@ int main(int argc, char **argv)
   JMP_BUF back;
   int e, num;
   char *p;
-
 #ifdef PIKE_EXTRA_DEBUG
-#ifdef HAVE_SIGNAL
   if (sizeof(void *) == 8) {
     /* 64-bit Solaris 10 in Xenofarm fails with SIGPIPE.
      * Force a core dump.
@@ -314,7 +309,6 @@ int main(int argc, char **argv)
     signal(SIGPIPE, abort);
   }
 #endif
-#endif
 
 #ifdef HAVE_MALLOPT
   TRACE((stderr, "Init malloc...\n"));
@@ -382,7 +376,7 @@ int main(int argc, char **argv)
 
   TRACE((stderr, "init_pike()\n"));
 
-  init_pike(argv, master_location + CONSTANT_STRLEN(MASTER_COOKIE));
+  init_pike(argv, master_file_location);
 
   for(e=1; e<argc; e++)
   {
@@ -441,13 +435,13 @@ int main(int argc, char **argv)
 		p++;
 	      }
 #ifdef _REENTRANT
-	      thread_stack_size=STRTOL(p,&p,0);
+	      thread_stack_size=strtol(p,&p,0);
 #endif
 	      p+=strlen(p);
 	      break;
 	    }
 	  }
-	  Pike_stack_size=STRTOL(p,&p,0);
+	  Pike_stack_size=strtol(p,&p,0);
 	  p+=strlen(p);
 
 	  if(Pike_stack_size < 256)
@@ -470,7 +464,7 @@ int main(int argc, char **argv)
 	  }else{
 	    p++;
 	  }
-	  set_pike_evaluator_limit(STRTOL(p, &p, 0));
+	  set_pike_evaluator_limit(strtol(p, &p, 0));
 	  p+=strlen(p);
 	  break;
 
@@ -480,14 +474,14 @@ int main(int argc, char **argv)
 	  {
 	    case '0': case '1': case '2': case '3': case '4':
 	    case '5': case '6': case '7': case '8': case '9':
-	      d_flag+=STRTOL(p+1,&p,10);
+	      d_flag+=strtol(p+1,&p,10);
 	      break;
 
 	    case 'c':
 	      p++;
-#if defined(YYDEBUG) || defined(PIKE_DEBUG)
+#if (defined(YYDEBUG) && (YYDEBUG==1)) && defined(PIKE_DEBUG)
 	      yydebug++;
-#endif /* YYDEBUG || PIKE_DEBUG */
+#endif /* YYDEBUG && PIKE_DEBUG */
 	      break;
 
 	    case 's':
@@ -550,7 +544,7 @@ int main(int argc, char **argv)
 
 	case 'a':
 	  if(p[1]>='0' && p[1]<='9')
-	    a_flag+=STRTOL(p+1,&p,10);
+	    a_flag+=strtol(p+1,&p,10);
 	  else
 	    a_flag++,p++;
 	  break;
@@ -560,7 +554,7 @@ int main(int argc, char **argv)
 	  switch (p[1]) {
 	    case '0': case '1': case '2': case '3': case '4':
 	    case '5': case '6': case '7': case '8': case '9':
-	      Pike_interpreter.trace_level+=STRTOL(p+1,&p,10);
+	      Pike_interpreter.trace_level+=strtol(p+1,&p,10);
 	      break;
 
 	    case 'g':
@@ -583,7 +577,7 @@ int main(int argc, char **argv)
 	    p+=strlen(p);
 	  }else{
 	    if(p[1]>='0' && p[1]<='9')
-	      p_flag+=STRTOL(p+1,&p,10);
+	      p_flag+=strtol(p+1,&p,10);
 	    else
 	      p_flag++,p++;
 	  }
@@ -591,7 +585,7 @@ int main(int argc, char **argv)
 
 	case 'l':
 	  if(p[1]>='0' && p[1]<='9')
-	    l_flag+=STRTOL(p+1,&p,10);
+	    l_flag+=strtol(p+1,&p,10);
 	  else
 	    l_flag++,p++;
 	  break;
@@ -637,7 +631,7 @@ int main(int argc, char **argv)
    *       the MASTER_COOKIE string in the binary.
    */
   add_pike_string_constant("__master_cookie",
-			   master_location, CONSTANT_STRLEN(MASTER_COOKIE));
+			   _master_location, CONSTANT_STRLEN(MASTER_COOKIE));
 
   if(SETJMP(back))
   {
@@ -645,7 +639,7 @@ int main(int argc, char **argv)
     {
       num=throw_value.u.integer;
     }else{
-      if (throw_value.type == T_OBJECT &&
+      if (TYPEOF(throw_value) == T_OBJECT &&
 	  throw_value.u.object->prog == master_load_error_program &&
 	  !get_master()) {
 	/* Report this specific error in a nice way. Since there's no
@@ -658,11 +652,10 @@ int main(int argc, char **argv)
 
 	move_svalue (Pike_sp++, &throw_value);
 	mark_free_svalue (&throw_value);
-	err = (struct generic_error_struct *)
+	err = 
 	  get_storage (Pike_sp[-1].u.object, generic_error_program);
 
-	t.type = PIKE_T_STRING;
-	t.u.string = err->error_message;
+	SET_SVAL(t, PIKE_T_STRING, 0, string, err->error_message);
 
 	init_buf(&buf);
 	describe_svalue(&t,0,0);
