@@ -797,7 +797,10 @@ static void call_destroy(struct object *o, enum object_destruct_reason reason)
 	  UNSETJMP (jmp);
 	  move_svalue (&throw_value, &err);
 	}
-	call_handle_error();
+	if (!SETJMP(jmp)) {
+	  call_handle_error();
+	}
+	UNSETJMP(jmp);
       }
 
       else {
@@ -2723,7 +2726,7 @@ static void f_magic_index(INT32 args)
     default:
     case 3:
       if (TYPEOF(sp[2-args]) != T_INT)
-	SIMPLE_BAD_ARG_ERROR ("::`->", 2, "void|int");
+	SIMPLE_BAD_ARG_ERROR ("::`->", 3, "void|int");
       type = sp[2-args].u.integer & 1;
       /* FALL THROUGH */
     case 2:
@@ -2945,7 +2948,7 @@ static void f_magic_indices (INT32 args)
     case 2:
       if (TYPEOF(sp[1-args]) != T_INT)
 	SIMPLE_BAD_ARG_ERROR ("::_indices", 2, "void|int");
-      type = sp[-args].u.integer;
+      type = sp[1-args].u.integer;
       /* FALL THROUGH */
     case 1:
       if (TYPEOF(sp[-args]) == T_INT) {
@@ -2958,13 +2961,13 @@ static void f_magic_indices (INT32 args)
 	    Pike_error("Object is destructed.\n");
 	  inherit = obj->prog->inherits + 0;
 	}
-      } else if (TYPEOF(sp[2-args]) == T_OBJECT) {
-	obj = sp[2-args].u.object;
+      } else if (TYPEOF(sp[-args]) == T_OBJECT) {
+	obj = sp[-args].u.object;
 	if (obj != MAGIC_THIS->o)
 	  Pike_error("::_indices context is not the current object.\n");
 	if(!obj->prog)
 	  Pike_error("::_indices on destructed object.\n");
-	inherit = obj->prog->inherits + SUBTYPEOF(sp[2-args]);
+	inherit = obj->prog->inherits + SUBTYPEOF(sp[-args]);
       } else {
 	SIMPLE_BAD_ARG_ERROR ("::_indices", 1, "void|object|int");
       }
@@ -3055,7 +3058,7 @@ static void f_magic_values (INT32 args)
     case 2:
       if (TYPEOF(sp[1-args]) != T_INT)
 	SIMPLE_BAD_ARG_ERROR ("::_indices", 2, "void|int");
-      type = sp[-args].u.integer;
+      type = sp[1-args].u.integer;
       /* FALL THROUGH */
     case 1:
       if (TYPEOF(sp[-args]) == T_INT) {
@@ -3068,13 +3071,13 @@ static void f_magic_values (INT32 args)
 	    Pike_error("Object is destructed.\n");
 	  inherit = obj->prog->inherits + 0;
 	}
-      } else if (TYPEOF(sp[2-args]) == T_OBJECT) {
-	obj = sp[2-args].u.object;
+      } else if (TYPEOF(sp[-args]) == T_OBJECT) {
+	obj = sp[-args].u.object;
 	if (obj != MAGIC_THIS->o)
 	  Pike_error("::_values context is not the current object.\n");
 	if(!obj->prog)
 	  Pike_error("::_values on destructed object.\n");
-	inherit = obj->prog->inherits + SUBTYPEOF(sp[2-args]);
+	inherit = obj->prog->inherits + SUBTYPEOF(sp[-args]);
       } else {
 	SIMPLE_BAD_ARG_ERROR ("::_values", 1, "void|object|int");
       }
@@ -3168,7 +3171,7 @@ static void f_magic_types (INT32 args)
     case 2:
       if (TYPEOF(sp[1-args]) != T_INT)
 	SIMPLE_BAD_ARG_ERROR ("::_types", 2, "void|int");
-      type = sp[-args].u.integer;
+      type = sp[1-args].u.integer;
       /* FALL THROUGH */
     case 1:
       if (TYPEOF(sp[-args]) == T_INT) {
@@ -3181,13 +3184,13 @@ static void f_magic_types (INT32 args)
 	    Pike_error("Object is destructed.\n");
 	  inherit = obj->prog->inherits + 0;
 	}
-      } else if (TYPEOF(sp[2-args]) == T_OBJECT) {
-	obj = sp[2-args].u.object;
+      } else if (TYPEOF(sp[-args]) == T_OBJECT) {
+	obj = sp[-args].u.object;
 	if (obj != MAGIC_THIS->o)
 	  Pike_error("::_types context is not the current object.\n");
 	if(!obj->prog)
 	  Pike_error("::_types on destructed object.\n");
-	inherit = obj->prog->inherits + SUBTYPEOF(sp[2-args]);
+	inherit = obj->prog->inherits + SUBTYPEOF(sp[-args]);
       } else {
 	SIMPLE_BAD_ARG_ERROR ("::_types", 1, "void|object|int");
       }
@@ -3263,6 +3266,9 @@ void init_object(void)
   offset=ADD_STORAGE(struct magic_index_struct);
   MAP_VARIABLE("__obj", tObj, ID_PROTECTED,
 	       offset + OFFSETOF(magic_index_struct, o), T_OBJECT);
+  /* NB: The type for this function is duplicated in las.c for
+   *     opcode F_MAGIC_INDEX.
+   */
   ADD_FUNCTION("`()", f_magic_index,
 	       tFunc(tStr tOr3(tVoid,tObj,tDeprecated(tInt)) tOr(tVoid,tInt),
 		     tMix), ID_PROTECTED);
@@ -3273,6 +3279,9 @@ void init_object(void)
   offset=ADD_STORAGE(struct magic_index_struct);
   MAP_VARIABLE("__obj", tObj, ID_PROTECTED,
 	       offset + OFFSETOF(magic_index_struct, o), T_OBJECT);
+  /* NB: The type for this function is duplicated in las.c for
+   *     opcode F_MAGIC_SET_INDEX.
+   */
   ADD_FUNCTION("`()", f_magic_set_index,
 	       tFunc(tStr tMix tOr3(tVoid,tObj,tDeprecated(tInt))
 		     tOr(tVoid,tInt), tVoid), ID_PROTECTED);
@@ -3283,6 +3292,9 @@ void init_object(void)
   offset=ADD_STORAGE(struct magic_index_struct);
   MAP_VARIABLE("__obj", tObj, ID_PROTECTED,
 	       offset + OFFSETOF(magic_index_struct, o), T_OBJECT);
+  /* NB: The type for this function is duplicated in las.c for
+   *     opcode F_MAGIC_INDICES.
+   */
   ADD_FUNCTION("`()", f_magic_indices,
 	       tFunc(tOr3(tVoid,tObj,tDeprecated(tInt)) tOr(tVoid,tInt),
 		     tArr(tStr)), ID_PROTECTED);
@@ -3293,6 +3305,9 @@ void init_object(void)
   offset=ADD_STORAGE(struct magic_index_struct);
   MAP_VARIABLE("__obj", tObj, ID_PROTECTED,
 	       offset + OFFSETOF(magic_index_struct, o), T_OBJECT);
+  /* NB: The type for this function is duplicated in las.c for
+   *     opcode F_MAGIC_VALUES.
+   */
   ADD_FUNCTION("`()", f_magic_values,
 	       tFunc(tOr3(tVoid,tObj,tDeprecated(tInt)) tOr(tVoid,tInt),
 		     tArray), ID_PROTECTED);
@@ -3303,9 +3318,12 @@ void init_object(void)
   offset=ADD_STORAGE(struct magic_index_struct);
   MAP_VARIABLE("__obj", tObj, ID_PROTECTED,
 	       offset + OFFSETOF(magic_index_struct, o), T_OBJECT);
+  /* NB: The type for this function is duplicated in las.c for
+   *     opcode F_MAGIC_TYPES.
+   */
   ADD_FUNCTION("`()", f_magic_types,
 	       tFunc(tOr3(tVoid,tObj,tDeprecated(tInt)) tOr(tVoid,tInt),
-		     tArray), ID_PROTECTED);
+		     tArr(tType(tMix))), ID_PROTECTED);
   magic_types_program=end_program();
 
   exit_compiler();
